/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaConvolution.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Handles NVIDIA Cuda memory interface.
**
*****************************************************************************/

#include "svlCudaConvolution.h"


svlCudaConvolution::svlCudaConvolution() {}

svlCudaConvolution::~svlCudaConvolution() {}

// Pad in width.
pair<int,int> svlCudaConvolution::halfPad(const float *patch, int pw, int ph, float* &pp)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = ph;
	if ( pw <= 4 ) {
		patch_w = 4;
	} else if ( pw <= 8 ) {
		patch_w = 8;
	} else if ( pw  <= 16 ) {
		patch_w = 16;
	} else {
		assert(false);
	}

	// Construct patch padded to above size.
	pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<ph; ++r ) {
		// Do on-patch columns
		for ( c=0; c<pw; ++c ) {
			*ptr++ =  patch[ pw*r + c ];
		}
		// Zeralize remainder columns
		for ( ; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
		}
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r ) {
		for ( c=0; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
		}
	}

#if 0
	// Dump patch for debugging.
	FILE *fo = fopen("patch.out","wb");
	fwrite(pp,sizeof(float),patch_w*patch_h,fo);
	fclose(fo);
#endif
	return pair<int,int>(patch_w,patch_h);
}

// The above with IplImage.
pair<int,int> svlCudaConvolution::halfPad(const IplImage *patch, float* &pp)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = patch->height;
	if ( patch->width <= 4 ) {
		patch_w = 4;
	} else if ( patch->width <= 8 ) {
		patch_w = 8;
	} else if ( patch->width  <= 16 ) {
		patch_w = 16;
	} else {
		assert(false);
	}

	// Construct patch padded to above size.
	pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c ) {
			*ptr++ =  CV_IMAGE_ELEM(patch, float, r, c);
		}
		// Zeralize remainder columns
		for ( ; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
		}
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r ) {
		for ( c=0; c<patch_w; ++c ) {
			*ptr++ = 0.0f;
		}
	}

	return pair<int,int>(patch_w,patch_h);
}

// Pad in width to the left.
pair<int,int> svlCudaConvolution::halfPadLeft(const float *patch, int pw, int ph, float* &pp)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = ph;
	if ( pw <= 4 ) {
		patch_w = 4;
	} else if ( pw <= 8 ) {
		patch_w = 8;
	} else if ( pw  <= 16 ) {
		patch_w = 16;
	} else {
		assert(false);
	}
	int buf = patch_w - pw;

	// Construct patch padded to above size.
	pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<ph; ++r ) {
		// Buffer column.
		for ( c=0; c<buf; ++c )
			*ptr++ = 0.0f;
		for ( c=buf; c<patch_w; ++c )
			*ptr++ =  patch[ pw*r + c - buf ];
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r )
		for ( c=0; c<patch_w; ++c )
			*ptr++ = 0.0f;

#if 0
	// Dump patch for debugging.
	FILE *fo = fopen("patch.out","wb");
	fwrite(pp,sizeof(float),patch_w*patch_h,fo);
	fclose(fo);
#endif
	return pair<int,int>(patch_w,patch_h);
}
// The above with an IplImage.
pair<int,int> svlCudaConvolution::halfPadLeft(const IplImage *patch, float* &pp)
{
	// Bin patch into a proper size and then convert to linear memory.
	int patch_w = -1, patch_h = patch->height;
	if ( patch->width <= 4 ) {
		patch_w = 4;
	} else if ( patch->width <= 8 ) {
		patch_w = 8;
	} else if ( patch->width  <= 16 ) {
		patch_w = 16;
	} else {
		assert(false);
	}
	int buf = patch_w - patch->width;

	// Construct patch padded to above size.
	pp = (float*)malloc(sizeof(float)*patch_w*patch_h);
	float *ptr = pp;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Buffer column.
		for ( c=0; c<buf; ++c )
			*ptr++ = 0.0f;
		for ( c=buf; c<patch_w; ++c )
			*ptr++ = CV_IMAGE_ELEM(patch, float, r, c-buf);
	}
	// Zeralize remainder rows
	for ( ; r<patch_h; ++r )
		for ( c=0; c<patch_w; ++c )
			*ptr++ = 0.0f;

	return pair<int,int>(patch_w,patch_h);
}

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
//void svlCudaConvolution::ccoeffNormed(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out,
//								   int img_w, int img_h, float patch_mean, float patch_std, svlCudaCoveringGrid *cudagrid)
//{
//	float *pp, *pmask;
//	pair<int,int> wh = fullPad(patch, pp, pmask);
//	int patch_w = wh.first, patch_h = wh.second;
//	// Run convolution.
//	svlCudaCcoeffNormed2d(*pitch_in, *pitch_out, pp, pmask, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h, cudagrid);
//	free(pp);
//	free(pmask);
//}

// Get padded patch size.
pair<int,int> svlCudaConvolution::paddedPatchSize(IplImage *patch)
{
	pair<int,int> ret;
	ret.second = min(16, patch->height);
	if ( patch->width <= 4 ) {
		ret.first = 4;
	} else if ( patch->width <= 8 ) {
		ret.first = 8;
	} else if ( patch->width <= 16 ) {
		ret.first = 16;
	} else {
		assert(false);
	}
	return ret;
}

// Pad the patch to device convolution size and store it and its mask in the provided memory.
void svlCudaConvolution::padPatchWithMask(IplImage *patch, float *mem)
{
	// Bin patch into a proper size and then convert to linear memory.
	pair<int,int> pad_wh = paddedPatchSize(patch);
	int sz = pad_wh.first * pad_wh.second;

	// Construct patch padded to above size.
	float *pdata = (float*)malloc(sizeof(float)*sz*2);
	float *ptr = pdata, *ptrm = pdata + sz;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<patch->height; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c ) {
			if ( patch->depth == IPL_DEPTH_8U ) {
				*ptr++ = (float)CV_IMAGE_ELEM(patch, unsigned char, r, c);
			} else if ( patch->depth == IPL_DEPTH_32F ) {
				*ptr++ = CV_IMAGE_ELEM(patch, float, r, c);
			}
			*ptrm++ = 1.0f;
		}
		// Zeralize remainder columns
		for ( ; c<pad_wh.first; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}
	// Zeralize remainder rows
	for ( ; r<pad_wh.second; ++r ) {
		for ( c=0; c<pad_wh.first; ++c ) {
			*ptr++ = 0.0f;
			*ptrm++ = 0.0f;
		}
	}
	memcpy(mem, pdata, sizeof(float)*sz*2);
	free(pdata);
}

// Pad the patch to device convolution size and store it in the provided memory.
void svlCudaConvolution::padPatchOnly(IplImage *patch, float *mem)
{
	// Bin patch into a proper size and then convert to linear memory.
	pair<int,int> pad_wh = paddedPatchSize(patch);
	int sz = pad_wh.first * pad_wh.second;

	// Construct patch padded to above size.
	float *pdata = (float*)malloc(sizeof(float)*sz);
	float *ptr = pdata;
	int r, c;
	// Do on-patch rows
	for ( r=0; r<pad_wh.second; ++r ) {
		// Do on-patch columns
		for ( c=0; c<patch->width; ++c ) {
			if ( patch->depth == IPL_DEPTH_8U ) {
				*ptr++ = (float)CV_IMAGE_ELEM(patch, unsigned char, r, c);
			} else if ( patch->depth == IPL_DEPTH_32F ) {
				*ptr++ = CV_IMAGE_ELEM(patch, float, r, c);
			}
		}
		// Zeralize remainder columns
		for ( ; c<pad_wh.first; ++c )
			*ptr++ = 0.0f;
	}
	// Zeralize remainder rows
	for ( ; r<pad_wh.second; ++r )
		for ( c=0; c<pad_wh.first; ++c )
			*ptr++ = 0.0f;
	
	memcpy(mem, pdata, sizeof(float)*sz);
	free(pdata);
}

// Performs CUDA_TM_CCOEFF_NORMED with direct summation over the image to compute the image statistics.
void svlCudaConvolution::ccoeffNormed(const svlCudaPitch *pin, const float *patch, int pw, int ph, svlCudaPitch *pout, int img_w, int img_h)
{
	float *pp;
	pair<int,int> wh = halfPad(patch, pw, ph, pp);
	const int patch_w = wh.first, patch_h = wh.second, N = pw*ph;

	// Create mask.
	int psz = patch_w*patch_h;
	float *pmask = (float*)malloc(sizeof(float)*psz);
	memset(pmask, 0, sizeof(float)*psz);
	for ( int r=0; r<ph; ++r )
		for ( int c=0; c<pw; ++c )
			pmask[ r*patch_w + c ] = 1.0f;

	// Compute mean and standard deviation.
	float mean = 0.0f, std = 0.0f;
	for ( int i=0; i<psz; ++i ) {
		mean += pp[i];
		std += pp[i] * pp[i];
	}
	mean /= float(N);
	std = sqrtf( std - float(N)*mean*mean );

	// Run convolution.
	svlCudaCcoeffNormed2d_v2(pin, pout, pp, pmask, patch_w, patch_h, N, mean, std, img_w, img_h);

	free(pp);
	free(pmask);
}

// The above with IplImage input.
void svlCudaConvolution::ccoeffNormed(const svlCudaPitch *pin, const IplImage *patch, svlCudaPitch *pout, int img_w, int img_h)
{
	float *pp;
	pair<int,int> wh = halfPad(patch, pp);
	const int patch_w = wh.first, patch_h = wh.second, N = patch->width*patch->height;

	// Create mask.
	int psz = patch_w*patch_h;
	float *pmask = (float*)malloc(sizeof(float)*psz);
	memset(pmask, 0, sizeof(float)*psz);
	for ( int r=0; r<patch->height; ++r )
		for ( int c=0; c<patch->width; ++c )
			pmask[ r*patch_w + c ] = 1.0f;

	// Compute mean and standard deviation.
	float mean = 0.0f, std = 0.0f;
	for ( int i=0; i<psz; ++i ) {
		mean += pp[i];
		std += pp[i] * pp[i];
	}
	mean /= float(N);
	std = sqrtf( std - float(N)*mean*mean );

	// Run convolution.
	svlCudaCcoeffNormed2d_v2(pin, pout, pp, pmask, patch_w, patch_h, N, mean, std, img_w, img_h);

	free(pp);
	free(pmask);
}

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
void svlCudaConvolution::cudaMatchTemplate(const svlCudaPitch *pitch_in, const float *patch, int pw, int ph, svlCudaPitch *pitch_out, int options, int img_w, int img_h)
{
	float *pp;
	pair<int,int> wh;
	switch ( options ) {
		case CUDA_TM_CCORR:
		case CUDA_TM_CCORR_SAME: // 'Same'-style convolution:
			wh = halfPad(patch, pw, ph, pp);
			svlCudaPlainConv2d(pitch_in, pitch_out, pp, wh.first, wh.second, img_w, img_h);
			break;
		case CUDA_TM_CCORR_FULL: // 'Full'-style convolution:
			wh = halfPadLeft(patch, pw, ph, pp);
			svlCudaPlainConv2d_full(pitch_in, pitch_out, pp, wh.first, wh.second, img_w, img_h);
			break;
		case CUDA_TM_CCOEFF_NORMED:
		case CUDA_TM_CCOEFF_NORMED_SAME: // Normalized correlation coefficient:
			ccoeffNormed(pitch_in, patch, pw, ph, pitch_out, img_w, img_h);
			break;
		default:
			printf("svlCudaConvolution::cudaMatchTemplate -- Invalid options: %d\n", options); assert(false);
	}
	free(pp);
}

// The above but accepting an IplImage for the patch.
void svlCudaConvolution::cudaMatchTemplate(const svlCudaPitch *pitch_in, const IplImage *patch, svlCudaPitch *pitch_out, int options, int img_w, int img_h)
{
	float *pp;
	pair<int,int> wh;
	switch ( options ) {
		case CUDA_TM_CCORR:
		case CUDA_TM_CCORR_SAME: // 'Same'-style convolution:
			wh = halfPad(patch, pp);
			svlCudaPlainConv2d(pitch_in, pitch_out, pp, wh.first, wh.second, img_w, img_h);
			break;
		case CUDA_TM_CCORR_FULL: // 'Full'-style convolution:
			wh = halfPadLeft(patch, pp);
			svlCudaPlainConv2d_full(pitch_in, pitch_out, pp, wh.first, wh.second, img_w, img_h);
			break;
		case CUDA_TM_CCOEFF_NORMED:
		case CUDA_TM_CCOEFF_NORMED_SAME: // Normalized correlation coefficient:
			ccoeffNormed(pitch_in, patch, pitch_out, img_w, img_h);
			break;
		default:
			printf("svlCudaConvolution::cudaMatchTemplate -- Invalid options: %d\n", options); assert(false);
	}
	free(pp);
}

// 3d pitch version of above.
void svlCudaConvolution::cudaMatchTemplate(const svlCudaPitch3d *pin, const float *patch, int pw, int ph, svlCudaPitch3d *pout, int options, int img_w, int img_h)
{
	for ( int i=0; i<pin->z; ++i ) {
		svlCudaPitch pin2 = pin->getPitch(i);
		svlCudaPitch pout2 = pout->getPitch(i);
		svlCudaConvolution::cudaMatchTemplate(&pin2, patch, pw, ph, &pout2, options, img_w, img_h);
	}
}
// And with IplImages:
void svlCudaConvolution::cudaMatchTemplate(const svlCudaPitch3d *pin, const IplImage *patch, svlCudaPitch3d *pout, int options, int img_w, int img_h)
{
	for ( int i=0; i<pin->z; ++i ) {
		svlCudaPitch pin2 = pin->getPitch(i);
		svlCudaPitch pout2 = pout->getPitch(i);
		svlCudaConvolution::cudaMatchTemplate(&pin2, patch, &pout2, options, img_w, img_h);
	}
}



// The above, but use v3.
// Performs CV_TM_CCOEFF_NORMED using row-integral images.
//void svlCudaConvolution::ccoeffNormed(const svlCudaPitch *pin, const IplImage *patch, svlCudaPitch *pout, svlCudaPitch *integ, svlCudaPitch *integ2, int img_w, int img_h)
//{
//	float *pp, *pmask;
//	pair<int,int> wh = halfPad(patch, pp, pmask);
//	int patch_w = wh.first, patch_h = wh.second;
//
//	// Run convolution.
//	svlCudaCcoeffNormed2d_v3(pin, pout, pp, pmask, integ, integ2, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h);
//	free(pp);
//	free(pmask);
//}

// Convolve a pitch with a patch described as an IplImage and output to another pitch.
//void svlCudaConvolution::normCCorr2(const svlCudaPitch *pitch_in, const IplImage *patch,
//									svlCudaPitch *pitch_out, svlCudaPitch *temp1, svlCudaPitch *temp2,
//									int img_w, int img_h, float patch_mean, float patch_std)
//{
//	float *pp, *pmask;
//	pair<int,int> wh = fullPad(patch, pp, pmask);
//	int patch_w = wh.first, patch_h = wh.second;
//
//	// Run convolution.
//	svlCudaCcoeffNormed2d_twostep(*pitch_in, *pitch_out, *temp1, *temp2,
//		pp, pmask, patch_w, patch_h, patch->height*patch->width, patch_mean, patch_std, img_w, img_h);
//	free(pp);
//	free(pmask);
//}
